#!/bin/bash

apt-mark auto \
    7zip \
    7zip-rar \
    acl \
    ant \
    apache2 \
    apache2-bin \
    aria2 \
    azure-cli \
    brotli \
    buildah \
    chrony \
    cifs-utils \
    clang-format-16 \
    clang-format-17 \
    clang-format-18 \
    clang-tidy-16 \
    clang-tidy-17 \
    clang-tidy-18 \
    cloud-init \
    dbus \
    dotnet-sdk-8.0 \
    eatmydata \
    efibootmgr \
    firefox \
    fonts-noto-color-emoji \
    gfortran \
    gfortran-12 \
    gfortran-13 \
    gfortran-14 \
    google-chrome-stable \
    google-cloud-cli \
    haveged \
    keyutils \
    kubectl \
    libeatmydata1 \
    libgmp-dev \
    liblz4-dev \
    libmysqlclient-dev \
    libpq-dev \
    libsqlite3-dev \
    libssl-dev \
    libtss2-esys-3.0.2-0t64 \
    libtss2-mu-4.0.1-0t64 \
    libtss2-rc0t64 \
    libwrap0 \
    libyaml-dev \
    lsscsi \
    lz4 \
    mediainfo \
    mercurial \
    microsoft-edge-stable \
    mysql-client \
    mysql-server \
    ncurses-base \
    ncurses-bin \
    ncurses-term \
    nginx \
    nvme-cli \
    packages-microsoft-prod \
    php-pear \
    php8.3 \
    php8.3-amqp \
    php8.3-apcu \
    php8.3-bcmath \
    php8.3-bz2 \
    php8.3-cgi \
    php8.3-cli \
    php8.3-common \
    php8.3-curl \
    php8.3-dba \
    php8.3-dev \
    php8.3-enchant \
    php8.3-fpm \
    php8.3-gd \
    php8.3-gmp \
    php8.3-igbinary \
    php8.3-imagick \
    php8.3-imap \
    php8.3-interbase \
    php8.3-intl \
    php8.3-ldap \
    php8.3-mbstring \
    php8.3-memcache \
    php8.3-memcached \
    php8.3-mongodb \
    php8.3-mysql \
    php8.3-odbc \
    php8.3-opcache \
    php8.3-pcov \
    php8.3-pgsql \
    php8.3-phpdbg \
    php8.3-pspell \
    php8.3-readline \
    php8.3-redis \
    php8.3-snmp \
    php8.3-soap \
    php8.3-sqlite3 \
    php8.3-sybase \
    php8.3-tidy \
    php8.3-xdebug \
    php8.3-xml \
    php8.3-xsl \
    php8.3-yaml \
    php8.3-zip \
    php8.3-zmq \
    pigz \
    podman \
    pollinate \
    postgresql-16 \
    powershell \
    python-babel-localedata \
    python3-babel \
    python3-dev \
    python3-jinja2 \
    python3-json-pointer \
    python3-jsonpatch \
    python3-jsonschema \
    python3-markupsafe \
    python3-parted \
    python3-pyrsistent \
    python3-serial \
    python3-six \
    python3-tz \
    python3-venv \
    rpm \
    ruby-full \
    session-manager-plugin \
    shellcheck \
    shim-signed \
    skopeo \
    snap \
    snmp \
    software-properties-common \
    sphinxsearch \
    sqlite3 \
    swig \
    sysstat \
    systemd-coredump \
    temurin-11-jdk \
    temurin-17-jdk \
    temurin-21-jdk \
    temurin-8-jdk \
    texinfo \
    tk \
    tree \
    tzdata \
    ubuntu-server \
    ubuntu-standard \
    unzip \
    upx-ucl \
    util-linux \
    walinuxagent \
    xvfb \
    xz-utils \
    zip \
    zsync

apt-get autoremove -y
